package be.iminds.ilabt.jfed.git;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

/**
 * Authentication setup for git repos.
 *
 *
 */
public interface GitAuthPreferences {
    enum GitAuthMethod { SSH, USER_PASS, OAUTH }
    interface GitAuth {
        @Nonnull GitAuthMethod getMethod();
        @Nonnull String getRepoUrl();
    }
    abstract class AbstractGitAuth implements GitAuth {
        @Nonnull private final GitAuthMethod method;
        @Nonnull private final String repoUrl;

        public AbstractGitAuth(@Nonnull GitAuthMethod method, @Nonnull String repoUrl) {
            this.method = method;
            this.repoUrl = repoUrl;
        }

        @Override
        @Nonnull public GitAuthMethod getMethod() {
            return method;
        }

        @Override
        @Nonnull public String getRepoUrl() {
            return repoUrl;
        }
    }
    class GitAuthSsh extends AbstractGitAuth {
        @Nonnull private final String privateKey;

        public GitAuthSsh(@Nonnull String repoUrl, @Nonnull String privateKey) {
            super(GitAuthMethod.SSH, repoUrl);
            this.privateKey = privateKey;
        }

        @Nonnull public String getPrivateKey() {
            return privateKey;
        }
    }
    class GitAuthUserPass extends AbstractGitAuth {
        @Nonnull private final String username;
        @Nullable private final String password;

        public GitAuthUserPass(@Nonnull String repoUrl, @Nonnull String username, @Nullable String password) {
            super(GitAuthMethod.USER_PASS, repoUrl);
            this.username = username;
            this.password = password;
        }

        @Nonnull public String getUsername() {
            return username;
        }

        @Nullable public String getPassword() {
            return password;
        }
    }

    /**
     *
     * @param repoUrl the URL of the repo
     * @return the auth method(s) supported by this URL
     */
    @Nonnull default Set<GitAuthMethod> getRepoUrlGitAuthMethod(@Nonnull String repoUrl) {
        //a bit simplistic, but good enough for now
        if (repoUrl.startsWith("https://"))
            return EnumSet.of(GitAuthMethod.USER_PASS, GitAuthMethod.OAUTH);
        if (repoUrl.startsWith("git@") ||
                (repoUrl.contains("@") && repoUrl.endsWith(".git")))
            return Collections.singleton(GitAuthMethod.SSH);
        throw new IllegalArgumentException("Unsupported repo url '"+repoUrl+"'");
    }

    /**
     *
     * @param repoUrl the URL of the repo
     * @return the auth method(s) available in this GitAuthPreferences for the repo
     */
    @Nonnull List<GitAuthMethod> getGitAuthMethodAvailable(String repoUrl);

    boolean hasAuth(String repoUrl);

    /**
     * @param allowedMethods optionally, specify the allowed methods, in order of preference  (if not specified, any method is returned)
     * @param repoUrl the URL of the repo
     * @return the git auth (or null if not present)  (notE: theoretically, the URL might change, if the auth available requires that)
     */
    @Nullable GitAuth getGitAuth(String repoUrl, GitAuthMethod... allowedMethods);

    @Nullable GitAuthSsh getGitAuthSsh(String repoUrl);
    @Nullable GitAuthUserPass getGitAuthUserPass(String repoUrl);
}
